/*
 * Playing field manages the GUI of the game, as well as managing the
 * moves. This class contains the state of the game and implements various
 * control methods: checkWin(), columnFull(), isFull(), countDiscs(). 
 */
package Game;

public class PlayingField {
    //Variables
    //size of the playing field
    private int width;
    private int height;
    //array to store all the discs
    public Disc discs[][];
    public int discsincol[];

    private int discammount; //stores the amount of discs in the playingfield
    private Player player1; //Variable to store instance player 1
    private Player player2; //Variable to store instance player 2
    private Player turn = player1; //Points to player 1 or player 2 to denote turn
    private Disc lastMove = null;
    private Disc[] winningSequence = null;

    
    /**
     * Create a new playing field.
     * 
     * A playing field is made with the specified width, height and players. The turn is assigned to one of the two players.
     * 
     * @param w Width of the field.
     * @param h Height of the field.
     * @param p1 Player 1.
     * @param p2 player 2.
     * @param t Initial turn.
     */
    public PlayingField(int w, int h, Player p1, Player p2, Player t) {
        width = w;
        height = h;

        discs = new Disc[width][height];
	discsincol = new int[width];
        
        discammount = 0;

	this.player1 = p1;
	this.player2 = p2;
	this.turn = t;
    }

    /**
     * Get the width of the playing field.
     * @return width
     */
    public int getWidth() {
        return width;
    }

    /**
     * Get the height of the playing field.
     * @return height
     */
    public int getheight() {
        return height;
    }

    /**
     * Get the player who's turn it is.
     * @return turn
     */
    public Player getTurn() {
        return turn;
    }

    /**
     * Get the height of the specified column.
     * @param col The column.
     * @return The amount of discs in the specified column.
     */
    public int getColumnHeight(int col) {
        //for (int i = 0; i < height; i++) if (discs[col][i] == null) return i;
        //return height;
	return discsincol[col];
    }

    /**
     * Get the disc at the specified colum and row.
     * @param col The column.
     * @param row The row.
     * @return The disc at (column,row).
     */
    public Disc getDiscAt(int col, int row) {
        //Return null if disc is outside playing field.
        if (col < 0 || col >= width || row < 0 || row >= height) return null;
        return discs[col][row];
    }


    /**
     * Function to check if the specified column is completely filled with discs.
     * @param col The column.
     * @return True if the column is full, false otherwise.
     */
    public boolean isColumnFull(int col) {
        if (col < 0 || col >= width) return false;
	//a column is full if the top most position is taken
        //return discs[col][height - 1] != null;
	return discsincol[col] >= height;
    }


    /**
     * Function to check if the complete playing field is filled with discs.
     * @return True if the complete playing field is full, otherwise false.
     */
    public boolean isFull() {
        //return discammount == width * height; //discs.lengthq;
	//if one column is not full, the field is not full
	for (int i = 0; i < width; i++) if (!isColumnFull(i)) return false;
	return true;
    }

    /**
     * Function to check if the previous move was a winning move.
     * 
     * The method assumes that all previous moves were no winning moves.
     * 
     * @return If the move was a winning move, return the player of the disc. Otherwise, return null.
     */
    public Player checkWin() {
	if (lastMove == null) return null;
        int lcol = lastMove.getColumn();    //store the column of the last move
        int lrow = lastMove.getRow();       //store the row of the last move
        Player owner = lastMove.getOwner(); //store the owner of the last move
	Disc[] sequence = new Disc[4];

        int cons = 0;       //count the consecutive discs

        //Horizontal 4 on a row
        for (int x = lcol - 3; x <= lcol + 3; x++) {
	    Disc d = getDiscAt(x, lrow);
            if (d == null || d.getOwner() != owner) {
		cons = 0;
		continue;
	    }
	    sequence[cons] = d;
            cons++;
            if (cons == 4) {
		winningSequence = sequence;
                return owner;
            }
        }

        //Vertical 4 on a row
	cons = 0;
        for (int y = lrow - 3; y <= lrow; y++) {
	    Disc d = getDiscAt(lcol, y);
            if (d == null || d.getOwner() != owner) {
		cons = 0;
		continue;
	    }
	    sequence[cons] = d;
            cons++;
            if (cons == 4) {
		winningSequence = sequence;
                return owner;
            }
        }

        //Topleft to Bottomright diagonal
	cons = 0;
        for (int r = -3; r <= 3; r++) {
	    Disc d = getDiscAt(lcol + r, lrow - r);
            if (d == null || d.getOwner() != owner) {
		cons = 0;
		continue;
	    }
	    sequence[cons] = getDiscAt(lcol + r, lrow - r);
            cons++;
            if (cons == 4) {
		winningSequence = sequence;
                return owner;
            }
        }

        //Topright to Bottomleft diagonal
	cons = 0;
        for (int r = -3; r <= 3; r++) {
	    Disc d = getDiscAt(lcol - r, lrow - r);
            if (d == null || d.getOwner() != owner) {
		cons = 0;
		continue;
	    }
	    sequence[cons] = getDiscAt(lcol - r, lrow - r);
            cons++;
            if (cons == 4) {
		winningSequence = sequence;
                return owner;
            }
        }

	return null;
    }

    /*
     * Returns the discs that make the winning sequence if there is any,
     * and null otherwise. Call checkWin() before using this getter.
     */
    
    /**
     * Function to return an array of discs that are part of the winning sequence.
     * @return An array containing the discs that are part of the winning sequence.
     */
    public Disc[] getWinningSequence() {
	return winningSequence;
    }

    /**
     * Function to get the amount of discs in the playing field.
     * @return: Amount of discs in the playing field.
     */
    public int countDiscs() {
        return discammount;
    }

    /**
     * Pre: c < width && !isColumnFull(c)
     * Post: a new Disc instance has been created and discs[discammount] 
     * points to the latest disc
     */
    
    /**
     * Function to place a new disc in the playing field.
     * 
     * The function adds a new disc to the playing field. The variables that are 
     * related to the playing field are updated.
     * @param c The column.
     * @param player The owner of the player.
     * @return True if the disc is placed successfully, false otherwise.
     */
    public boolean putDisc(int c, Player player) {
        if (isColumnFull(c)) return false;
	int r = getColumnHeight(c);
	if (getDiscAt(c, r) != null) System.out.println("Disc position not empty!!! at " + Integer.toString(getDiscAt(c, r).getColumn()));
	if (discs[c][r] != null) System.out.println("Disc slot not empty!!!");
	lastMove = new Disc(c, r, player);
        //discs[discammount] = lastMove;
        discs[c][r] = lastMove;
	discsincol[c]++;
        discammount++;
	turn = (turn == player1) ? player2 : player1;
	return true;
    }

    /**
     * Get the disc of the previous move.
     * @return The most recent disc that was placed.
     */
    public Disc getLastMove() {
	return lastMove;
    }

    /**
     * Make a clone of the current playing field.
     * @return A clone of the playing field.
     */
    public PlayingField clone() {
	PlayingField f = new PlayingField(width, height, player1, player2, turn);
        for (int x = 0; x < width; x++) System.arraycopy(discs[x], 0, f.discs[x], 0, height);
	System.arraycopy(discsincol, 0, f.discsincol, 0, width);
	f.discammount = discammount;
	f.player1 = player1;
	f.player2 = player2;
	f.turn = turn;
	return f;
    }

    /**
     * Print the current field in the standard output.
     */
    public void printField() {
	for (int y = height - 1; y >= 0; y--) {
	    String s = "";
	    for (int x = 0; x < width; x++) {
		s += "[";
		if (discs[x][y] == null) s += " ";
		else s += discs[x][y].getOwner().color.substring(0, 1);
		s += "]";
	    }
	    System.out.println(s);
	}
    }
}